@@ -87,9 +87,9 @@ class Photo(Implicit, Persistent, PropertyManager, RoleManager, Item):
     security=ClassSecurityInfo()
 
     def __init__(self, id, title, file, content_type='', precondition='',
-                 store='Image', engine='ImageMagick', quality=75, pregen=0, timeout=0):
+                 store='Image', engine='PIL', quality=75, pregen=0, timeout=0):
 
-        self.__version__ = '1.2.3'
+        self.__version__ = '1.2.4'
         self.id = id
         self.title = title
 
@@ -344,14 +344,17 @@ class Photo(Implicit, Persistent, PropertyManager, RoleManager, Item):
     # Photo processing
     #
 
-    def _resize(self, display, width, height, engine='ImageMagick', quality=75):
+    def _resize(self, display, width, height, engine='PIL', quality=75):
         """Resize and resample photo."""
         origimg = self._original
         newimg = StringIO()
         if engine == 'PIL':  # Use PIL
             img = PIL.Image.open(origimg._PILdata())
             fmt = img.format
-            img = img.resize((width, height))
+            try:
+                img = img.resize((width, height), PIL.Image.ANTIALIAS)
+            except AttributeError:
+                img = img.resize((width, height))
             img.save(newimg, fmt, quality=quality)
         elif engine == 'ImageMagick':  # Use ImageMagick
             if sys.platform == 'win32':
@@ -404,16 +407,18 @@ class Photo(Implicit, Persistent, PropertyManager, RoleManager, Item):
         """Create all displays."""
         for display in self._displays.keys():
             self._makeDisplayPhoto(display)
-            
+
     def _getAspectRatioSize(self, width, height):
         """Return proportional dimensions within desired size."""
-        img_width, img_height = (self.width(), self.height())
-        if height > img_height * width / img_width:
-            height = img_height * width / img_width
+        img_width, img_height = self.width(), self.height()
+        sw = float(width) / img_width
+        sh = float(height) / img_height
+        if sw <= sh:
+            height = int(sw * img_height + 0.5)
         else:
-            width =  img_width * height / img_height
+            width = int(sh * img_width + 0.5)
         return (width, height)
-
+                                                                                
     def _validImage(self):
         """At least see if it *might* be valid."""
         return self._original._isValid()
@@ -462,12 +467,24 @@ class Photo(Implicit, Persistent, PropertyManager, RoleManager, Item):
                 self._makeDisplayPhotos()
         else:
             # Adding a new Photo.
-            # At this point, the object is not yet in its final context.
-            # Since ExtImage needs to know the URL of the new Photo object
-            # so it can determine where to create new files, we store
+            # In Zope 2.7.3 and before, the object is not yet in its final
+            # context. Since ExtImage needs to know the URL of the new Photo
+            # object so it can determine where to create new files, we store
             # the image data and let manage_afterAdd() generate the displays
             # once the Photo is in its final context.
             self._data = file.read()
+
+            # Zope 2.7.4 and after does _setObject before calling this PUT(),
+            # so we can do all the work now.
+            if self.absolute_url(1):
+                store = self.propertysheets.get('photoconf').getProperty('store')
+                if store == 'Image': from PhotoImage import PhotoImage
+                elif store == 'ExtImage': from ExtPhotoImage import PhotoImage
+                self._original = PhotoImage(self.id, self.title, path=self.absolute_url(1))
+                self._original.manage_upload(StringIO(self._data), self.content_type())
+                delattr(self, '_data')
+                if self._validImage():
+                    self._makeDisplayPhotos()
             
         RESPONSE.setStatus(204)
         return RESPONSE
@@ -481,10 +498,6 @@ class Photo(Implicit, Persistent, PropertyManager, RoleManager, Item):
         """Handle STAT requests."""
         return self._original.manage_FTPstat(REQUEST)
 
-    def manage_FTPlist(self, REQUEST):
-        """Handle LIST requests."""
-        return self._original.manage_FTPlist(REQUEST)
-
     #
     # Management Interface
     #
@@ -613,9 +626,10 @@ class Photo(Implicit, Persistent, PropertyManager, RoleManager, Item):
 
     def manage_afterAdd(self, item, container):
         """Handle pasting of new photos."""
-        if not hasattr(self, '_original'):
-            # Added Photo (vs. imported)
-            # See note in PUT()
+
+        # For Zope 2.7.3 and before.
+        # See notes in PUT()
+        if hasattr(self, '_data'):
             store = self.propertysheets.get('photoconf').getProperty('store')
             if store == 'Image': from PhotoImage import PhotoImage
             elif store == 'ExtImage': from ExtPhotoImage import PhotoImage
@@ -625,10 +639,10 @@ class Photo(Implicit, Persistent, PropertyManager, RoleManager, Item):
             if self._validImage():
                 self._makeDisplayPhotos()
 
-        self._original.manage_afterAdd(item, container)
-        if hasattr(self, '_photos'):
-            for photo in self._photos.values():
-                photo.manage_afterAdd(item, container)
+            self._original.manage_afterAdd(item, container)
+            if hasattr(self, '_photos'):
+                for photo in self._photos.values():
+                    photo.manage_afterAdd(item, container)
 
     def manage_beforeDelete(self, item, container):
         """Delete (mark for undo) each photo file."""
@@ -644,14 +658,17 @@ manage_addPhotoForm = DTMLFile('dtml/addPhotoForm', globals())
 
 def manage_addPhoto(dispatcher, id, title, file,
                     content_type='', precondition='',
-                    store='Image', engine='ImageMagick', quality=75,
+                    store='Image', engine='PIL', quality=75,
                     timeout=0, pregen=0, REQUEST=None):
     """Add Photo object."""
     id = cookId(id, title, file)[0]
     dest = dispatcher.Destination()
     photo = Photo(id, title, '', content_type, precondition,
                   store, engine, quality, pregen, timeout)
-    photo._data = file.read()
+    try:
+        photo._data = file.read()
+    except AttributeError:
+        photo._data = str(file)
     dest._setObject(id, photo)
     # Images are generated at this point by manage_afterAdd()
     
@@ -660,15 +677,23 @@ def manage_addPhoto(dispatcher, id, title, file,
     parent = self.aq_parent
     if parent.meta_type == 'Photo Folder':
         props = parent.propertysheets.get('photos')
-        for propid, value in props.propertyItems():
-            try: self.manage_addProperty(propid, value, props.getPropertyType(propid))
-            except: pass
+        for prop in props.propertyMap():
+            try:
+                if prop['type'] in ('selection', 'multiple selection'):
+                    self.manage_addProperty(prop['id'], prop['select_variable'], prop['type'])
+                    self.manage_changeProperties({prop['id']: props.getProperty(prop['id'])})
+                else:
+                    self.manage_addProperty(prop['id'], props.getProperty(prop['id']), prop['type'])
+            except: pass  # Properties may have been added automatically
         self._displays = parent._displays.copy()
 
     if REQUEST is not None:
         try:    url=dispatcher.DestinationURL()
         except: url=REQUEST['URL1']
-        REQUEST.RESPONSE.redirect('%s/manage_main' % url)
+        if REQUEST.form.get('action', None) == ' Add and Edit ':
+            REQUEST.RESPONSE.redirect('%s/%s/manage_workspace' % (url, id))
+        else:
+            REQUEST.RESPONSE.redirect('%s/manage_workspace' % url)
     return id
 
 InitializeClass(Photo)
